UNIT Queue2ADT;
(* Queue2ADT implements the abstract data type FIFO queue. *)

INTERFACE

USES
  Prog3Data;

TYPE
  QueuePtrType = ^QueueNodeType;
  QueueNodeType = RECORD
    Info : QueueElementType;  (* the queue user's data      *)
    Next : QueuePtrType       (* pointer to next queue node *)
  END;  (* QueueNodeType *)
  QueueType = RECORD
    Front : QueuePtrType;  (* accesses front of queue *)
    Rear  : QueuePtrType   (* accesses rear of queue  *)
  END; (* QueueType *)
(*********************************************************)

(***********************************************************)

PROCEDURE CreateQueue
  (VAR Queue : QueueType);
  (* Function: Initializes Queue to empty state. *)  
  (* Pre:  None                                  *)
  (* Post: Queue is empty.                       *)

(***********************************************************)

PROCEDURE DestroyQueue
  (VAR Queue : QueueType);
  (* Function: Removes all elements from Queue,  *)
  (*   leaving the queue empty.                  *)
  (* Pre:  Queue has been created.               *)
  (* Post: Queue is empty.                       *)

(***********************************************************)

FUNCTION EmptyQueue
  (Queue : QueueType) : Boolean;
  (* Function: Indicates whether Queue is empty. *)
  (* Pre:  Queue has been created.               *)
  (* Post: EmptyQueue = (Queue is empty)         *)

(***********************************************************)

FUNCTION FullQueue
  (Queue : QueueType) : Boolean;
  (* Function: Indicates whether Queue is full. *)
  (* Pre:  Queue has been created.              *)
  (* Post: FullQueue = (Queue is full)          *)

(***********************************************************)

PROCEDURE Enqueue
  (VAR Queue  : QueueType;
   NewElement : QueueElementType);
  (* Function: Adds NewElement to the rear of Queue. *)
  (* Pre:  Queue has been created and is not full.   *)
  (* Post: Queue = original Queue with NewElement    *)
  (*       added to rear.                            *)

(***********************************************************)

PROCEDURE Dequeue
  (VAR Queue      : QueueType;
   VAR DeqElement : QueueElementType);
  (* Function: Removes front element from Queue and     *)
  (*   returns its value as DeqElement.                 *)
  (* Pre:  Queue has been created and is not empty.     *)
  (* Post: Queue = original Queue with front element    *)
  (*         removed.                                   *)
  (*       DeqElement = front element of original Queue *)

(*********************************************************)

PROCEDURE PrintQueue
  (    Queue : QueueType;
   VAR Output : Text);
  {* Purpose : To print out the data for a queue         *}
  {*     Pre : Queue has been created                    *}
  {*    Post : All values in queue are outputed          *}

(*********************************************************)
(*-------------------------------------------------------*)
(*********************************************************)

(*********************************************************)
(*-------------------------------------------------------*)
(*********************************************************)

IMPLEMENTATION

PROCEDURE CreateQueue
  (VAR Queue : QueueType);
  (* Pre:  None                                *)
  (* Post: Queue.Rear and Queue.Front are NIL. *)

BEGIN (* CreateQueue *)
  Queue.Front := NIL;
  Queue.Rear  := NIL
END; (* CreateQueue *)
(**********************************************************)

FUNCTION EmptyQueue
  (Queue : QueueType) : Boolean;
  (* Pre:  Queue has been initialized.     *)
  (* Post:  EmptyQueue = Queue.Front = NIL *)

BEGIN (* EmptyQueue *)
  EmptyQueue := (Queue.Front = NIL)
END; (* EmptyQueue *)

(*********************************************************)

FUNCTION FullQueue
  (Queue : QueueType) : Boolean;
  (* Pre:  Queue has been initialized.  *)
  (* Post: FullQueue = False            *)

BEGIN (* FullQueue *)
  (* This function is not meaningful for this queue *)
  (* implementation, so just return False.          *)
  FullQueue := False
END; (* FullQueue *)

(*********************************************************)

PROCEDURE Enqueue
  (VAR Queue  : QueueType;
   NewElement : QueueElementType);
  (* Pre:  Queue has been initialized.                      *)
  (*       NOT FullQueue(Queue)                             *)
  (* Post: Queue.Rear points to node containing NewElement  *)
  (*         in its Info field and NIL in its Next field.   *)
  (*       If original queue was empty, Queue.Front points  *)
  (*         to Queue.Rear.                                 *)
VAR
  NewNode : QueuePtrType;

BEGIN (* Enqueue *)
  (* Create a node for the new element. *)
  New (NewNode);
  NewNode^.Info := NewElement;
  NewNode^.Next := NIL;
  (* Insert new node at rear of Queue. *)
  IF Queue.Rear = NIL (* queue is empty *)
    THEN Queue.Front := NewNode
    ELSE Queue.Rear^.Next := NewNode;
  (* Update external pointer to rear of Queue. *)
  Queue.Rear := NewNode
END; (* Enqueue *)

(***************************************************)

PROCEDURE Dequeue
  (VAR Queue      : QueueType;
   VAR DeqElement : QueueElementType);
  (* Pre:  Queue has been initialized.                   *)
  (*       NOT EmptyQueue(Queue)                         *)
  (* Post: Set DeqElement to Info field of Queue.Front^. *)
  (*       Remove and dispose Queue.Front.               *)
  (*       If EmptyQueue(Queue), set Queue.Rear to NIL.  *)
VAR
  TempPtr : QueuePtrType;
BEGIN (* Dequeue *)
  TempPtr := Queue.Front; (* save it for Dispose *)
  (* Take DeqElement from front node of Queue. *)
  DeqElement := Queue.Front^.Info;
  (* Remove front node from Queue. *)
  Queue.Front := Queue.Front^.Next;
  IF Queue.Front = NIL (* queue is now empty *)
    THEN Queue.Rear := NIL;
  Dispose (TempPtr)
END;  (* Dequeue *)

(**********************************************************)

PROCEDURE DestroyQueue
  (VAR Queue : QueueType);
  (* Pre:  Queue has been initialized.         *)
  (* Post: Queue elements have been disposed.  *)
  (*       Queue.Front and Queue.Rear are NIL. *)

VAR
  TempPtr : QueuePtrType;

BEGIN (* DestroyQueue *)
  WHILE Queue.Front <> NIL DO
    BEGIN
      TempPtr := Queue.Front;
      Queue.Front := Queue.Front^.Next;
      Dispose (TempPtr)
    END; (* WHILE *)
  Queue.Rear := NIL
END;  (* DestroyQueue *)

(**********************************************************)
PROCEDURE PrintQueue
  (    Queue : QueueType;
   VAR Output : Text);
  {* Purpose : To print out the data for a queue         *}
  {*     Pre : Queue has been created                    *}
  {*    Post : All values in queue are outputed          *}

VAR
  TempPtr : QueuePtrType;   {* Temp pointer to traverse *}

BEGIN {* PrintQueue *}

  {* Start at beginning of the queue *}
  TempPtr := Queue.Front;

  {* Traverse the queue until the end of it and print ID Nums *}
  WHILE TempPtr <> NIL DO
    BEGIN {* While TempPtr *}
      write (Output, TempPtr^.Info.IDNum, ' ');
      TempPtr := TempPtr^.Next;
    END; {* While TempPtr *}

END; {* PrintQueue *}

(**********************************************************)

END.  (* unit Queue2ADT *)
