PROGRAM Solitaire (Input, Output);

(* This program is a simulation of a card game. *)
(* See text for rules of the game               *)

CONST
  DeckSize = 52;
  HalfDeck = 26;

TYPE
  NodePointer = ^NodeType;
  Suits = (Club, Diamond, Heart, Spade);
  CardType = RECORD
               Suit : Suits;
               Rank : 1..13              (* 1 is ace; 13 is king *)
             END; (* Record *)
  NodeType = RECORD
               Card : CardType;
               Link : NodePointer
             END; (* Record *)

VAR
  Deck,                (* External pointer to deck of cards *)
  OnTable,             (* NodeTypes face up on the table    *)
  DiscardPile:         (* NodeTypes on discard pile         *)
    NodePointer;
  NumberOfShuffles,    (* Number of shuffles per game       *)
  NumberOfGames,       (* Number of games to play           *)
  GamesPlayed,         (* Number of games played            *)
  GamesWon:            (* Number of games won               *)
    Integer;
  Seed:                (* Used with random number generator *)
    Real;

(****************************************************************
****)

PROCEDURE CreateDeck (VAR Deck:         (* Pointer to list of cards
*)
                            NodePointer);

(* Creates a list of DeckSize nodes.  Each node represents a     *)
(* playing card.  Deck is the external pointer to the list.  The *)
(* cards are in order by suit and by rank *)

VAR
  NewNode,          (* Pointer to newest card   *)
  Current:          (* Pointer to previous card *)
    NodePointer;
  Count:            (* Loop control variable    *)
    Integer;

BEGIN (* CreateDeck *)
  New(Deck);
  Deck^.Card.Suit := Club;
  Deck^.Card.Rank := 1;
  Current := Deck;
  (* Loop to create balance of deck *)
  FOR Count := 2 to DeckSize DO
    BEGIN
      New(NewNode);
      (* Test for change of suit *)
      IF Current^.Card.Rank = 13
        THEN
          (* Change suit *)
          BEGIN
            NewNode^.Card.Suit := Succ(Current^.Card.Suit);
            NewNode^.Card.Rank := 1
          END
        ELSE
          (* Increment rank *)
          BEGIN
            NewNode^.Card.Suit := Current^.Card.Suit;
            NewNode^.Card.Rank := Current^.Card.Rank + 1
          END;
      Current^.Link := NewNode;
      Current := NewNode
    END;
  (* Set end-of-list marker *)
  Current^.Link := NIL
END;  (* CreateDeck *)

(****************************************************************
****)

PROCEDURE RemoveTop (VAR List:          (* Pointer to list of cards
*)
                           NodePointer;
                     VAR Element:       (* Top card returned      
 *)
                           CardType);

(* The first card (node) is removed from List. The entire node is
*)
(* returned in Element *)

VAR
  Pointer:           (* Temporary pointer *)
    NodePointer;

BEGIN (* RemoveTop *)
  IF List <> NIL
    THEN
      BEGIN
        Pointer := List;
        Element := List^.Card;
        List := List^.Link;
        Dispose(Pointer)
      END
END;  (* RemoveTop *)

(****************************************************************
****)

PROCEDURE InsertTop (VAR List:          (* Pointer to list of cards
*)
                           NodePointer;
                         Element:       (* Type of card to insert 
 *)
                           CardType);

(* Element is inserted as the first card (node) in List. *)
(* Assumption:  List has been initialized to NIL         *)

VAR
  Pointer:           (* Temporary pointer *)
    NodePointer;

BEGIN (* InsertTop *)
  New(Pointer);
  Pointer^.Card := Element;
  Pointer^.Link := List;
  List := Pointer
END;  (* InsertTop *)

(****************************************************************
****)

PROCEDURE Shuffle (VAR Deck:            (* Pointer to list of cards 
*)
                         NodePointer;
                       NumberOfShuffles: (* Times to perform
shuffle *)
                         Integer);

(* Shuffle takes an input list of DeckSize nodes and returns the *)
(* list with the nodes in a different order.  The input list is  *)
(* divided into two parts, which are then merged. The process is *)
(* repeated NumberOfShuffles times *)

VAR
  HalfA,             (* Half of the list      *)
  HalfB:             (* Half of the list      *)
    NodePointer;
  TempCard:          (* Temporary card        *)
    CardType;
  Count1,            (* Loop control variable *)
  Count2:            (* Loop control variable *)
    Integer;
  RanNumber:         (* Random real number    *)
    Real;
  SizeOfCut:         (* Size of HalfA         *)
    0..DeckSize;

(****************************************************************
****)

  PROCEDURE Random (VAR Seed, RanNumber:  (* Seed for generator   
 *)
                          Real);          (* Random generated value
*)

  CONST
    Pi = 3.14159;

  VAR
    Temp:        (* Temporary variable for computing random number
*)
      Real;

  BEGIN (* Random *)
    Temp := Seed + Pi;
    Temp := Exp(5.0 * Ln(Temp));
    Seed := Temp - Trunc(Temp);
    RanNumber := Seed
  END;  (* Random *)

(****************************************************************
****)

  PROCEDURE Merge (    ShorterList,     (* Pointer to list of cards
*)
                       LongerList:      (* Pointer to list of cards
*)
                         NodePointer;
                       Length:          (* Length of shorter list 
 *)
                         Integer;
                   VAR Deck:            (* Resulting list of cards 
*)
                         NodePointer);

  (* ShorterList and LongerList are merged into Deck.  Length is *)
  (* number of nodes in ShorterList *)

  VAR
    Count:                   (* Loop control variable *)
      Integer;
    TempCard:                (* Temporary card        *)
      CardType;

  BEGIN (* Merge *)
    (* Merge two halves back into Deck *)
    FOR Count := 1 TO Length DO
      BEGIN
        RemoveTop(ShorterList, TempCard);
        InsertTop(Deck, TempCard);
        RemoveTop(LongerList, TempCard);
        InsertTop(Deck, TempCard)
      END;
    FOR Count := 1 TO (DeckSize - 2 * Length) DO
      BEGIN
        RemoveTop(LongerList, TempCard);
        InsertTop(Deck, TempCard)
      END
  END;  (* Merge *)

(****************************************************************
****)

BEGIN (* Shuffle *)
  FOR Count1 := 1 TO NumberOfShuffles DO
    BEGIN
      Random(Seed, RanNumber);
      SizeOfCut := (Trunc(RanNumber * 100.0) MOD DeckSize) + 1;
      HalfA := NIL;
      HalfB := NIL;
      (* Divide Deck into two parts *)
      FOR Count2 := 1 TO SizeOfCut DO
        BEGIN
          RemoveTop(Deck, TempCard);
          InsertTop(HalfA, TempCard)
        END;
      FOR Count2 := SizeOfCut + 1 TO DeckSize DO
        BEGIN
          RemoveTop(Deck, TempCard);
          InsertTop(HalfB, TempCard)
        END;
      IF SizeOfCut <= HalfDeck
        THEN
          Merge(HalfA, HalfB, SizeOfCut, Deck)
        ELSE
          Merge(HalfB, HalfA, DeckSize - SizeOfCut, Deck)
    END
END;  (* Shuffle *)

(****************************************************************
****)

FUNCTION First (OnTable:                (* Pointer to list of cards
*)
                  NodePointer):
                    Suits;

(* First returns the suit of the first card (node) in OnTable *)

BEGIN (* First *)
  First := OnTable^.Card.Suit
END;  (* First *)

(****************************************************************
****)

FUNCTION Second (OnTable:               (* Pointer to list of cards
*)
                   NodePointer):
                     Suits;

(* Second returns the suit of the second card in OnTable *)

BEGIN (* Second *)
  Second := OnTable^.Link^.Card.Suit
END;  (* Second *)

(****************************************************************
****)

FUNCTION Third (OnTable:                (* Pointer to list of cards
*)
                  NodePointer):
                    Suits;

(* Third returns the suit of the third card in OnTable *)

BEGIN (* Third *)
  Third := OnTable^.Link^.Link^.Card.Suit
END;  (* Third *)

(****************************************************************
****)

FUNCTION Fourth (OnTable:               (* Pointer to list of cards
*)
                   NodePointer):
                     Suits;

(* Fourth returns the suit of the fourth card in OnTable *)

BEGIN (* Fourth *)
  Fourth := OnTable^.Link^.Link^.Link^.Card.Suit
END;  (* Fourth *)

(****************************************************************
****)

PROCEDURE MoveFour (VAR OnTable,        (* Pointer to list of cards
*)
                        DiscardPile:    (* Pointer to list of cards
*)
                          NodePointer);

(* The first four cards are moved from OnTable to DiscardPile *)

VAR
  TempCard:        (* Temporary card        *)
    CardType;
  Count:           (* Loop control variable *)
    Integer;

BEGIN (* MoveFour *)
  FOR Count := 1 TO 4 DO
    BEGIN
      RemoveTop(OnTable, TempCard);
      InsertTop(DiscardPile, TempCard)
    END
END;  (* MoveFour *)

(****************************************************************
****)

PROCEDURE MoveTwo (VAR OnTable,         (* Pointer to list of cards
*)
                       DiscardPile:     (* Pointer to list of cards
*)
                         NodePointer);

(* The second and third cards are moved from OnTable to DiscardPile
*)

VAR
  TempCard,       (* Temporary card *)
  First:          (* Temporary card *)
    CardType;

BEGIN (* MoveTwo *)
  RemoveTop(OnTable, First);
  (* Remove second card *)
  RemoveTop(OnTable, TempCard);
  InsertTop(DiscardPile, TempCard);
  (* Remove third card *)
  RemoveTop(OnTable, TempCard);
  InsertTop(DiscardPile, TempCard);
  (* Replace first card *)
  InsertTop(OnTable, First)
END;  (* MoveTwo *)

(****************************************************************
****)

FUNCTION FourOnTable (OnTable:          (* Pointer to list of cards
*)
                        NodePointer):
                          Boolean;

(* FourOnTable is True if OnTable contains at least four cards *)

VAR
  Count:         (* Loop control variable *)
    Integer;

BEGIN (* FourOnTable *)
  Count := 0;
  WHILE (Count <> 4) AND (OnTable <> NIL) DO
    BEGIN
      Count := Count + 1;
      OnTable := OnTable^.Link
    END;
  FourOnTable := (Count = 4)
END;  (* FourOnTable *)

(****************************************************************
****)

PROCEDURE TryRemove (VAR OnTable,       (* Pointer to list of cards
*)
                         DiscardPile:   (* Pointer to list of cards
*)
                           NodePointer);

(* If first (top) four cards are the same suit, they are moved from
*)
(* OnTable to DiscardPile.  If the first card and the fourth card 
 *)
(* are the same suit, the second and third cards are moved from   
 *)
(* OnTable to DiscardPile.  Process continues until no further move
*)
(* can be made *)

VAR
  Moved:         (* Flag to record whether a move has been made *)
    Boolean;

BEGIN (* TryRemove *)
  Moved := True;
  WHILE FourOnTable(OnTable) AND Moved DO
    IF First(OnTable) = Fourth(OnTable)
      THEN                                   (* A move will be made
*)
        BEGIN
          IF (First(OnTable) = Second(OnTable)) AND
             (First(OnTable) = Third(OnTable))
            THEN
              (* Four alike *)
              MoveFour(OnTable, DiscardPile)
            ELSE
              (* First and fourth alike *)
              MoveTwo(OnTable, DiscardPile)
        END
      ELSE
        Moved := False
END;  (* TryRemove *)
(****************************************************************
****)

PROCEDURE PlayGame (VAR Deck,             (* Deck of playing cards 
*)
                        OnTable,          (* Cards face up on table
*)
                        DiscardPile:      (* Cards on discard pile 
*)
                          NodePointer);

(* Places the next card in the deck face up on the table *)
(* and calls TryRemove to apply rules for moving         *)

VAR
  Count:                        (* Loop control variable *)
    Integer;
  TempCard:                     (* Temporary card        *)
    CardType;

BEGIN (* PlayGame *)
  FOR Count := 1 to DeckSize DO
    BEGIN
      (* Turn up card *)
      RemoveTop(Deck, TempCard);
      InsertTop(OnTable, TempCard);
      TryRemove(OnTable, DiscardPile)
    END
END;  (* PlayGame *)

(****************************************************************
****)

PROCEDURE Recreate (VAR Deck,             (* Deck of playing cards 
*)
                         OnTable,         (* Cards face up on table
*)
                         DiscardPile:     (* Cards on discard pile 
*)
                         NodePointer);

(* Gathers cards and puts them back into Deck *)

VAR
  TempCard:       (* Temporary card *)
    CardType;

BEGIN (* Recreate *)
  (* Move cards from OnTable to Deck *)
  WHILE OnTable <> NIL DO
    BEGIN
      RemoveTop(OnTable, TempCard);
      InsertTop(Deck, Tempcard)
    END;
  (* Move cards from DiscardPile to Deck *)
  WHILE DiscardPile <> NIL DO
    BEGIN
      RemoveTop(DiscardPile, TempCard);
      InsertTop(Deck, TempCard)
    END
END;  (* Recreate *)

(****************************************************************
****)

BEGIN (* Solitaire *)
  (* Initialize *)
  Seed := 4.0;
  GamesWon := 0;
  OnTable := NIL;
  DiscardPile := NIL;
  Writeln('Enter number of games to play.');
  Readln(NumberOfGames);
  Writeln('Enter number of shuffles per game.');
  Readln(NumberOfShuffles);
  CreateDeck(Deck);
  FOR GamesPlayed := 1 TO NumberOfGames DO
    BEGIN
      Shuffle(Deck, NumberOfShuffles);
      PlayGame(Deck, OnTable, DiscardPile);
      (* Determine if game was won *)
      IF OnTable = NIL
        THEN
          (* Game won *)
          GamesWon := GamesWon + 1;
      Recreate(Deck, OnTable, DiscardPile)
    END;
  Writeln('Number of games played: ', NumberOfGames);
  Writeln('Number of games won:    ', GamesWon);
  Readln;
END.  (* Solitaire *)
